// Rastrigin_GA.cpp : Defines the entry point for the console application.
//


#include <time.h>
#include <stdlib.h>
#include <mpi.h>
#include "helperFunctions.h"

int main(int argc, char **argv)
{
	//MPI Init logic
	int rank, size;
	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	printf("SIZE = %d RANK = %d\n",size,rank);
	MPI_Status status;
	MPI_Request request;

	//GA vars
	float min = -5.12f;
	float max = 5.12f;
	float mutationRate = 0.1f;
	unsigned numBits = 15;
	unsigned numGenerations = 1000;
	unsigned population_size = 8192;
	unsigned dimension = 20;
	unsigned interval = 20;
	unsigned longevity = 20;
	char e;

	float *memory;
	memory = new float[population_size * dimension];

	float *fitnessValues;
	fitnessValues = new float[population_size];

	unsigned *chromosomes;
	chromosomes = new unsigned[population_size * dimension];

	unsigned *random_cross_values;
	random_cross_values = new unsigned[population_size * dimension];

	unsigned *random_mutation_values;
	random_mutation_values = new unsigned[population_size * dimension];

	unsigned *firstParent, *secondParent;
	firstParent = new unsigned[dimension];
	secondParent = new unsigned[dimension];
	float firstValue, secondValue;

	float max_int_value = (float)(pow((float)2,(float)numBits) - 1);

	srand ((unsigned)(time(NULL)+rank));
	
	for (unsigned i = 0; i < population_size; i++) {
		for (unsigned j = 0; j < dimension; j++)
		{
		chromosomes[i * dimension + j] = (unsigned)((float)(rand()/ (float)RAND_MAX)*max_int_value);
		random_cross_values[i * dimension + j] = (unsigned)rand();
		random_mutation_values[i * dimension + j] = (unsigned)(rand() % (dimension + 1));
		chromosomeToFloats(&chromosomes[i * dimension], &memory[i * dimension], dimension, min, max, max_int_value);
		}
		fitnessValues[i] = rastrigin(&memory[i * dimension], dimension);
    }
	dataCopy(&chromosomes[0], &firstParent[0], dimension);
	dataCopy(&chromosomes[dimension], &secondParent[0], dimension);
	firstValue = fitnessValues[0];
	secondValue = fitnessValues[1];

	
	clock_t elapsedTicks = clock();
    printf("Start\n");

	for(int i = 0; i < numGenerations; i++)
	{
		for(int k = 0; k < population_size; k++)
		{
			fitnessValues[k] = rastrigin(&memory[k*dimension], dimension);
		}
		selection(&fitnessValues[0], &chromosomes[0], &firstParent[0], &secondParent[0], &firstValue, &secondValue, population_size, dimension);
		for(int k = 0; k < population_size; k++)
		{
			crossover(&firstParent[0], &secondParent[0], &chromosomes[k*dimension], &random_cross_values[k*dimension], dimension);
			
			chromosomeToFloats(&chromosomes[k*dimension], &memory[k * dimension], dimension, min, max, max_int_value);
		}
	}

	elapsedTicks = clock() - elapsedTicks;
	double elapsedTime = (elapsedTicks)/(CLOCKS_PER_SEC/1000);
	printf("\n%f\n", elapsedTime);
	displayDataFloat(&firstValue, 1, 1);

	

	delete[] memory;
	delete[] chromosomes;
	delete[] random_cross_values;
	delete[] random_mutation_values;
	delete[] fitnessValues;
	delete[] firstParent;
	delete[] secondParent;
	e = getchar();
	MPI_Finalize();
	return 0;
}

