#include "openclUtils.h"
#include <mpi.h>
#include <time.h>
#include "helperFunctions.h"
#include "mpiUtils.h"
#include "float.h"

#define MAX_SOURCE_SIZE (0x100000) 

int main(int argc, char **argv) 
{ 
	//MPI Init logic
	int rank, size;
	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	printf("SIZE = %d RANK = %d\n",size,rank);
	int sync_tag = 0;
	int chromosome_tag = 10;
	int request_tag = 2;
	int finished_tag = 3;
	int server = 0;
	int longevity = 20;
	int interval = 20;

	//GA vars
	float min = -5.12f;
	float max = 5.12f;
	float mutationRate = 0.1f;
	unsigned numBits = 4;
	unsigned numGenerations = 1000;
	unsigned population_size = 256;
	unsigned dimension = 20;
			

	float max_int_value = (float)(pow((float)2,(float)numBits) - 1);

	float *firstParentFloats;
	firstParentFloats = new float[dimension];

	if (rank == server)
	{
		int clients = size - 1;
		printf("Running server\n");
		unsigned *data;
		data = new unsigned[size * dimension];
		float *values_req;
		values_req = new float[size];
		bool *requests;
		requests = new bool[size];
		for (int i = 0; i < size; i++)
		{
			requests[i] = false;
			values_req[i] = FLT_MAX;
		}
		float *chromosomeFloats;
		chromosomeFloats = new float[dimension];
		int numOfFinished = 0;
		float someValue;
		while (true)
		{
			//sync pokretanja
			for (int i = 1; i <= clients; i++)
			{
				if (RecievedMessage(i, sync_tag))
				{
					AcceptRequest(i, sync_tag);
					numOfFinished++; 
				}
			}
			if (numOfFinished == clients) 
			{
				for (int i = 1; i <= clients; i++)
				{
					SendRequest(i, sync_tag, &someValue);
				}
				goto next;
			}
		}
next:
		numOfFinished = 0;
		while (true)
		{
			//provjera kraja izvoenja
			for (int i = 1; i <= clients; i++)
			{
				if (RecievedMessage(i, finished_tag))
				{
					values_req[i] = AcceptRequest(i, finished_tag);
					numOfFinished++;
				}
			}
			if (numOfFinished == clients) 
			{
				displayDataFloat(&values_req[1], 1, clients);
				break;
			}

			//hvatanje zahtjeva za jedinkama i odgavaranje na iste
			for (int i = 1; i <= clients; i++)
			{
				if (RecievedMessage(i, request_tag))
				{
					//dobiven zahtjev za jedinkom
					AcceptRequest(i, request_tag);
					requests[i] = true;
					for(int j = 1; j <= clients; j++)
					{
						if ((i!=j) &&  (values_req[j] < values_req[i]))
						{
							SendChromosome(&data[j*dimension], dimension, i, chromosome_tag + i);
						}
					}
				}
				if (RecievedMessage(i, chromosome_tag))
				{
					AcceptChromosome(&data[i * dimension],dimension, i, chromosome_tag);
					chromosomeToFloats(&data[i * dimension], &firstParentFloats[0], dimension, min, max, max_int_value);
					values_req[i] = rastrigin(&firstParentFloats[0], dimension);
				}
			}

		}
	}
	else {
		
		printf("Running OpenCL platform code\n");

		float *memory;
		memory = new float[population_size * dimension];

		float *memory_b;
		memory_b = new float[population_size * dimension];

		float *fitnessValues;
		fitnessValues = new float[population_size];

		unsigned *chromosomes;
		chromosomes = new unsigned[population_size * dimension];

		unsigned *random_cross_values;
		random_cross_values = new unsigned[population_size * dimension];

		unsigned *random_mutation_values;
		random_mutation_values = new unsigned[population_size * dimension];

		unsigned *firstParent, *secondParent,*recvBuffer;
		firstParent = new unsigned[dimension];
		secondParent = new unsigned[dimension];
		recvBuffer = new unsigned[dimension];

		float firstValue, secondValue;

		//openCL vars
		cl_platform_id * platform_id = NULL;
		cl_device_id device_id = NULL;
		cl_context context = NULL;
		cl_command_queue command_queue = NULL;
		cl_mem d_chromosomes = NULL;
		cl_mem d_firstParent = NULL;
		cl_mem d_secondParent = NULL;
		cl_mem d_random_cross = NULL;
		cl_mem d_random_mutate = NULL;
		cl_mem d_dimension = NULL;
		cl_mem d_floatValues = NULL;
		cl_mem d_numBits = NULL;
		cl_mem d_min = NULL;
		cl_mem d_max = NULL;
		cl_mem d_max_int_value = NULL;
		cl_mem d_memory = NULL;
		cl_program program = NULL;
		cl_kernel k_crossover = NULL;
		cl_kernel k_calculateFloats = NULL;
		cl_kernel k_mutation = NULL;
		cl_uint ret_num_devices;
		cl_uint ret_num_platforms;
		cl_int ret;
		size_t size_of_population = population_size * dimension * sizeof(unsigned);
		size_t size_of_chromosome = dimension * sizeof(unsigned);
		size_t size_of_floatValues = population_size * dimension * sizeof(float);
		FILE *fp;
		const char fileName[] = "C:\\MPI_TEST\\kernels.cl";
		size_t source_size;
		char *source_str;

		/* Load kernel source code */
		fp = fopen(fileName, "r");
		if (fp) {
			source_str = (char *)malloc(MAX_SOURCE_SIZE);
			source_size = fread(source_str, 1, MAX_SOURCE_SIZE, fp);
			fclose(fp);    
		}
		/*Initialize Data */


		srand ((unsigned)(time(NULL)+rank));

		for (unsigned i = 0; i < population_size; i++) {
			for (unsigned j = 0; j < dimension; j++)
			{
				chromosomes[i * dimension + j] = (unsigned)((float)(rand()/ (float)RAND_MAX)*max_int_value);
				random_cross_values[i * dimension + j] = (unsigned)rand();
				random_mutation_values[i * dimension + j] = (unsigned)(rand() % (dimension + 1));
				chromosomeToFloats(&chromosomes[i * dimension], &memory[i * dimension], dimension, min, max, numBits);
			}
			fitnessValues[i] = rastrigin(&memory[i * dimension], dimension);
		}


		dataCopy(&chromosomes[0], &firstParent[0], dimension);
		dataCopy(&chromosomes[dimension], &secondParent[0], dimension);
		firstValue = fitnessValues[0];
		secondValue = fitnessValues[1];

		/* Get platform/device information */
		cl_uint numPlatforms;
		ret = clGetPlatformIDs(0, NULL, &numPlatforms);
		platform_id = (cl_platform_id *)alloca(sizeof(cl_platform_id) * numPlatforms);
		for (int i = 0; i < numPlatforms; i++)
		{
			if (i == rank % numPlatforms)
			{
				ret = clGetPlatformIDs(numPlatforms, platform_id, NULL); //imamo cl_platform_id
				ret = clGetDeviceIDs(platform_id[i], CL_DEVICE_TYPE_DEFAULT, 1, &device_id, &ret_num_devices);
				DisplayInfo(platform_id[i], CL_PLATFORM_NAME);
			}
		}

		context = clCreateContext(NULL, 1, &device_id, NULL, NULL, &ret);
		command_queue = clCreateCommandQueue(context, device_id, NULL, &ret);
		//Otvaramo memoriju koja nam je potrebna za rad
		d_chromosomes = clCreateBuffer(context, CL_MEM_READ_WRITE, size_of_population, NULL, &ret); 
		d_firstParent = clCreateBuffer(context, CL_MEM_READ_ONLY, size_of_chromosome, NULL, &ret); 
		d_secondParent = clCreateBuffer(context, CL_MEM_READ_ONLY, size_of_chromosome, NULL, &ret); 
		d_random_cross = clCreateBuffer(context, CL_MEM_READ_WRITE, size_of_population, NULL, &ret); 
		d_random_mutate = clCreateBuffer(context, CL_MEM_READ_WRITE, size_of_population, NULL, &ret); 
		d_dimension = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(unsigned), NULL, &ret); 
		d_min = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(float), NULL, &ret); 
		d_max = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(float), NULL, &ret); 
		d_numBits = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(unsigned), NULL, &ret);
		d_max_int_value = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(float), NULL, &ret); 
		d_memory = clCreateBuffer(context, CL_MEM_READ_WRITE, population_size * dimension * sizeof(float), NULL, &ret);



		ret = clEnqueueWriteBuffer(command_queue, d_random_cross, CL_TRUE, 0, size_of_population, random_cross_values, 0, NULL, NULL); 
		ret = clEnqueueWriteBuffer(command_queue, d_random_mutate, CL_TRUE, 0, size_of_population, random_mutation_values, 0, NULL, NULL); 

		ret = clEnqueueWriteBuffer(command_queue, d_chromosomes, CL_TRUE, 0, size_of_population, chromosomes, 0, NULL, NULL); 
		ret = clEnqueueWriteBuffer(command_queue, d_dimension, CL_TRUE, 0, sizeof(unsigned), &dimension, 0, NULL, NULL); 
		ret = clEnqueueWriteBuffer(command_queue, d_min, CL_TRUE, 0, sizeof(float), &min, 0, NULL, NULL); 
		ret = clEnqueueWriteBuffer(command_queue, d_max, CL_TRUE, 0, sizeof(float), &max, 0, NULL, NULL);
		ret = clEnqueueWriteBuffer(command_queue, d_max_int_value, CL_TRUE, 0, sizeof(float), &max_int_value, 0, NULL, NULL); 
		ret = clEnqueueWriteBuffer(command_queue, d_memory, CL_TRUE, 0, sizeof(float) + dimension * population_size, memory, 0, NULL, NULL);

		//ON-LINE stvaranje kernela
		program = clCreateProgramWithSource(context, 1, (const char **)&source_str, (const size_t *)&source_size, &ret); 
		ret = clBuildProgram(program, 1, &device_id, NULL, NULL, NULL); 
		k_crossover = clCreateKernel(program, "crossover", &ret); 
		k_calculateFloats = clCreateKernel(program, "calculateFloats", &ret); 
		k_mutation = clCreateKernel(program, "mutation", &ret);

		//prosljeivnje parametara
		ret = clSetKernelArg(k_crossover, 0, sizeof(cl_mem), (void *)&d_firstParent); 
		ret = clSetKernelArg(k_crossover, 1, sizeof(cl_mem), (void *)&d_secondParent);
		ret = clSetKernelArg(k_crossover, 2, sizeof(cl_mem), (void *)&d_chromosomes);
		ret = clSetKernelArg(k_crossover, 3, sizeof(cl_mem), (void *)&d_random_cross);
		ret = clSetKernelArg(k_crossover, 4, sizeof(cl_mem), (void *)&d_dimension);

		ret = clSetKernelArg(k_calculateFloats, 0, sizeof(cl_mem), (void *)&d_chromosomes); 
		ret = clSetKernelArg(k_calculateFloats, 1, sizeof(cl_mem), (void *)&d_memory); 
		ret = clSetKernelArg(k_calculateFloats, 2, sizeof(cl_mem), (void *)&d_min); 
		ret = clSetKernelArg(k_calculateFloats, 3, sizeof(cl_mem), (void *)&d_max); 
		ret = clSetKernelArg(k_calculateFloats, 4, sizeof(cl_mem), (void *)&d_max_int_value); 

		ret = clSetKernelArg(k_mutation, 0, sizeof(cl_mem), (void *)&d_chromosomes); 
		ret = clSetKernelArg(k_mutation, 1, sizeof(cl_mem), (void *)&d_random_mutate); 

		size_t global_work_size[3] = {population_size * dimension, 0, 0};

		clFinish(command_queue);
		
		//sinkronizacija
		SendRequest(server, sync_tag, &firstValue);
		while(true)
		{
			if (RecievedMessage(server, sync_tag)) break;
		}
		clock_t elapsedTicks = clock();
		printf("Start\n");

		bool changed; //spremamo rezutat selekcije, true ako su pronaene nove jedinke
		
		
		int c_interval = interval;
		int c_longevity = longevity;

		float recvValue;
		for(int i = 0; i < numGenerations; i++)
		{
			for(int k = 0; k < population_size; k++)
			{
				//chromosomeToFloats(&chromosomes[k*dimension], &memory[k*dimension], dimension, min, max, max_int_value);
				fitnessValues[k] = rastrigin(&memory[k*dimension], dimension);
			}
			changed = selection(&fitnessValues[0], &chromosomes[0], &firstParent[0], &secondParent[0], &firstValue, &secondValue, population_size, dimension);
			
			//logika komunikacije sa serverom
			c_longevity--;
			c_interval--;
			if (changed) c_longevity = longevity;
			if (c_interval <=0)
			{
				c_interval = interval;
				SendChromosome(&firstParent[0], dimension, server, chromosome_tag);
			}
			if (c_longevity <=0)
			{
				c_longevity = longevity;
				SendRequest(server, request_tag, &firstValue);
			}
			if (RecievedMessage(server, chromosome_tag + rank))
			{
				AcceptChromosome(&recvBuffer[0], dimension, server, chromosome_tag + rank);
				chromosomeToFloats(&recvBuffer[0], &firstParentFloats[0], dimension, min, max, max_int_value);
				recvValue = rastrigin(&firstParentFloats[0], dimension);
				if (recvValue < firstValue)
				{
					firstValue = recvValue;
					dataCopy(&recvBuffer[0], &firstParent[0], dimension);
					changed = true;
				}
				else if (recvValue < secondValue)
				{
					secondValue = recvValue;
					dataCopy(&recvBuffer[0], &secondParent[0], dimension);
					changed = true;
				}					
			}


			ret = clEnqueueWriteBuffer(command_queue, d_firstParent, CL_TRUE, 0, size_of_chromosome, firstParent, 0, NULL, NULL); 
			ret = clEnqueueWriteBuffer(command_queue, d_secondParent, CL_TRUE, 0, size_of_chromosome, secondParent, 0, NULL, NULL); 


			for (unsigned g = 0; g < population_size; g++) 
			{
				for (unsigned h = 0; h < dimension; h++)
				{
					random_cross_values[g * dimension + h] = (unsigned)rand();
					random_mutation_values[g * dimension + h] = (unsigned)(rand() % (dimension + 1));
				}
			}
			ret = clEnqueueWriteBuffer(command_queue, d_random_cross, CL_TRUE, 0, size_of_population, random_cross_values, 0, NULL, NULL); 
			ret = clEnqueueWriteBuffer(command_queue, d_random_mutate, CL_TRUE, 0, size_of_population, random_mutation_values, 0, NULL, NULL); 

			ret = clEnqueueNDRangeKernel(command_queue, k_crossover, 1, NULL, global_work_size, NULL, 0, NULL, NULL);
			ret = clEnqueueNDRangeKernel(command_queue, k_mutation, 1, NULL, global_work_size, NULL, 0, NULL, NULL);
			ret = clEnqueueNDRangeKernel(command_queue, k_calculateFloats, 1, NULL, global_work_size, NULL, 0, NULL, NULL);
			ret = clEnqueueReadBuffer(command_queue, d_memory, CL_TRUE, 0, size_of_floatValues, memory, 0, NULL, NULL);

		}

		elapsedTicks = clock() - elapsedTicks;
		double elapsedTime = (elapsedTicks)/(CLOCKS_PER_SEC/1000);
		printf("\nVrijeme:%f\n\n", elapsedTime);

		/* Finalization */
		ret = clFlush(command_queue);
		ret = clFinish(command_queue);
		ret = clReleaseKernel(k_crossover);
		ret = clReleaseKernel(k_calculateFloats);
		ret = clReleaseKernel(k_mutation);
		ret = clReleaseProgram(program);
		ret = clReleaseMemObject(d_chromosomes);
		ret = clReleaseMemObject(d_memory);
		ret = clReleaseMemObject(d_min);
		ret = clReleaseMemObject(d_max);
		ret = clReleaseMemObject(d_random_cross);
		ret = clReleaseMemObject(d_random_mutate);
		ret = clReleaseMemObject(d_firstParent);
		ret = clReleaseMemObject(d_secondParent);


		ret = clReleaseCommandQueue(command_queue);
		ret = clReleaseContext(context);

		delete[] memory;
		delete[] chromosomes;
		delete[] random_cross_values;
		delete[] random_mutation_values;
		delete[] fitnessValues;
		delete[] firstParent;
		delete[] secondParent;

		free(source_str);
		SendRequest(server, finished_tag, &firstValue);
	}	
	if (rank == server) char e = getchar();
	MPI_Finalize();
	return 0;
}
